API_KEY_ENV_KEY_MAP = {
    "arbitrum": "ARBISCAN_API_KEY",
    "ethereum": "ETHERSCAN_API_KEY",
    "fantom": "FTMSCAN_API_KEY",
    "optimism": "OPTIMISTIC_ETHERSCAN_API_KEY",
    "polygon": "POLYGONSCAN_API_KEY",
    "avalanche": "SNOWTRACE_API_KEY",
    "bsc": "BSCSCAN_API_KEY",
}
